@@ -203,13 +203,10 @@ char *strip_last_component(const char *path);
 void utf8_init(void);
 bool using_utf8(void);
 #endif
-char *addstrings(char* str1, size_t len1, char* str2, size_t len2);
-bool is_byte(int c);
-bool is_alpha_mbchar(const char *c);
-bool is_blank_mbchar(const char *c);
-bool is_ascii_cntrl_char(int c);
-bool is_cntrl_mbchar(const char *c);
-bool is_word_mbchar(const char *c, bool allow_punct);
+bool is_alpha_char(const char *c);
+bool is_blank_char(const char *c);
+bool is_cntrl_char(const char *c);
+bool is_word_char(const char *c, bool allow_punct);
 char control_mbrep(const char *c, bool isdata);
 #ifdef ENABLE_UTF8
 int mbwidth(const char *c);
@@ -267,7 +264,6 @@ void copy_from_buffer(linestruct *somebuffer);
 void cut_marked(bool *right_side_up);
 #endif
 void do_snip(bool copying, bool marked, bool until_eof, bool append);
-bool is_cuttable(bool test_cliff);
 void cut_text(void);
 #ifndef NANO_TINY
 void copy_text(void);
@@ -279,13 +275,9 @@ void paste_text(void);
 /* Most functions in files.c. */
 void make_new_buffer(void);
 #ifndef NANO_TINY
-int delete_lockfile(const char *lockfilename);
+bool delete_lockfile(const char *lockfilename);
 #endif
 bool open_buffer(const char *filename, bool new_buffer);
-#ifdef ENABLE_SPELLER
-bool replace_buffer(const char *filename, undo_type action, bool marked,
-		const char *operation);
-#endif
 void set_modified(void);
 void prepare_for_display(void);
 #ifdef ENABLE_MULTIBUFFER
@@ -375,15 +367,15 @@ void do_page_down(void);
 #ifdef ENABLE_JUSTIFY
 void do_para_begin(linestruct **line);
 void do_para_end(linestruct **line);
-void do_para_begin_void(void);
-void do_para_end_void(void);
+void to_para_begin(void);
+void to_para_end(void);
 #endif
-void do_prev_block(void);
-void do_next_block(void);
+void to_prev_block(void);
+void to_next_block(void);
 void do_prev_word(bool allow_punct);
 bool do_next_word(bool after_ends, bool allow_punct);
-void do_prev_word_void(void);
-void do_next_word_void(void);
+void to_prev_word(void);
+void to_next_word(void);
 void do_home(void);
 void do_end(void);
 void do_up(void);
@@ -444,23 +436,9 @@ void confirm_margin(void);
 #endif
 void unbound_key(int code);
 bool okay_for_view(const keystruct *shortcut);
-void inject(char *output, size_t output_len, bool filtering);
+void inject(char *burst, size_t count);
 
 /* Most functions in prompt.c. */
-void inject_into_answer(int *the_input, size_t input_len, bool filtering);
-void do_statusbar_home(void);
-void do_statusbar_end(void);
-void do_statusbar_left(void);
-void do_statusbar_right(void);
-void do_statusbar_backspace(void);
-void do_statusbar_delete(void);
-void do_statusbar_cut_text(void);
-void do_statusbar_uncut_text(void);
-#ifndef NANO_TINY
-void do_statusbar_prev_word(void);
-void do_statusbar_next_word(void);
-#endif
-void do_statusbar_verbatim_input(void);
 size_t get_statusbar_page_start(size_t start_col, size_t column);
 void put_cursor_at_end_of_answer(void);
 void add_or_remove_pipe_symbol_from_answer(void);
@@ -524,9 +502,7 @@ void do_undo(void);
 void do_redo(void);
 void do_enter(void);
 #ifndef NANO_TINY
-RETSIGTYPE cancel_command(int signal);
-bool execute_command(const char *command);
-void discard_until(const undostruct *thisitem, openfilestruct *thefile, bool keep);
+void discard_until(const undostruct *thisitem);
 void add_undo(undo_type action, const char *message);
 void update_multiline_undo(ssize_t lineno, char *indentation);
 void update_undo(undo_type action);
@@ -596,6 +572,7 @@ void new_magicline(void);
 void remove_magicline(void);
 #endif
 #ifndef NANO_TINY
+bool mark_is_before_cursor(void);
 void get_region(const linestruct **top, size_t *top_x,
 		const linestruct **bot, size_t *bot_x, bool *right_side_up);
 void get_range(const linestruct **top, const linestruct **bot);
@@ -616,8 +593,7 @@ int parse_kbinput(WINDOW *win);
 int get_kbinput(WINDOW *win, bool showcursor);
 int get_byte_kbinput(int kbinput);
 int get_control_kbinput(int kbinput);
-int *get_verbatim_kbinput(WINDOW *win, size_t *kbinput_len);
-int *parse_verbatim_kbinput(WINDOW *win, size_t *count);
+char *get_verbatim_kbinput(WINDOW *win, size_t *count);
 #ifdef ENABLE_MOUSE
 int get_mouseinput(int *mouse_row, int *mouse_col, bool allow_shortcuts);
 #endif
